@@ -1,169 +1,17 @@
-/*++
-/* NAME
-/*	agetty 8
-/* SUMMARY
-/*	alternative System V/SunOS 4 getty
-/* SYSTEM V SYNOPSIS
-/*	agetty [-i] [-l login_program] [-m] [-t timeout] port baud_rate,...
-/* SUNOS 4 SYNOPSIS
-/*	agetty [-h] [-l login_program] [-m] [-t timeout] baud_rate,... port
-/* DESCRIPTION
-/*	\fIagetty\fP opens a tty port, prompts for a login name and invokes
-/*	the /bin/login command. It is normally invoked by \fIinit(8)\fP.
-/*
-/*	\fIagetty\fP has several \fInon-standard\fP features that are useful
-/*	for hard-wired and for dial-in lines:
-/* .IP o
-/*	Adapts the tty settings to parity bits and to erase, kill,
-/*	end-of-line and uppercase characters when it reads a login name.
-/*	The program can handle 7-bit characters with even, odd, none or space
-/*	parity, and 8-bit characters with no parity. The following special
-/*	characters are recognized: @ and Control-U (kill); #, DEL and
-/*	back space (erase); carriage return and line feed (end of line).
-/* .IP o
-/*	Optionally deduces the baud rate from the CONNECT messages produced by 
-/*	Hayes(tm)-compatible modems.
-/* .IP o
-/*	Optionally does not hang up when it is given an already opened line 
-/*	(useful for call-back applications).
-/* .IP o
-/*	Optionally does not display the contents of the \fI/etc/issue\fP file
-/*	(System V only).
-/* .IP o
-/*	Optionally invokes a non-standard login program instead of
-/*	\fI/bin/login\fP.
-/* .IP o
-/*	Optionally turns on hard-ware flow control (SunOS 4 only).
-/* .PP
-/*	This program does not use the \fI/etc/gettydefs\fP (System V) or
-/*	\fI/etc/gettytab\fP (SunOS 4) files.
-/* ARGUMENTS
-/* .fi
-/* .ad
-/* .TP
-/* port
-/*	A path name relative to the \fI/dev\fP directory. If a "-" is 
-/*	specified, \fIagetty\fP assumes that its standard input is 
-/*	already connected to a tty port and that a connection to a 
-/*	remote user has already been established. 
-/* .sp
-/*	Under System V, a "-" \fIport\fP argument should be preceded 
-/*	by a "--".
-/* .TP
-/* baud_rate,...
-/*	A comma-separated list of one or more baud rates. Each time 
-/*	\fIagetty\fP receives a BREAK character it advances through 
-/*	the list, which is treated as if it were circular. 
-/* .sp
-/*	Baud rates should be specified in descending order, so that the 
-/*	null character (Ctrl-@) can also be used for baud rate switching.
-/* OPTIONS
-/* .fi
-/* .ad
-/* .TP
-/* -h (SunOS 4 only)
-/*	Enable hardware (RTS/CTS) flow control. It is left up to the
-/*	application to disable software (XON/XOFF) flow protocol where
-/*	appropriate.
-/* .TP
-/* -i (System V only)
-/*	Do not display the contents of \fI/etc/issue\fP before writing the
-/*	login prompt. Terminals or communications hardware may become confused
-/*	when receiving lots of text at the wrong baud rate; dial-up scripts
-/*	may fail if the login prompt is preceded by too much text.
-/* .TP
-/* -l login_program
-/*	Invoke the specified \fIlogin_program\fP instead of /bin/login.
-/*	This allows the use of a non-standard login program (for example,
-/*	one that asks for a dial-up password or that uses a different 
-/*	password file).
-/* .TP
-/* -m
-/*	Try to extract the baud rate the \fIconnect\fP status message 
-/*	produced by some Hayes(tm)-compatible modems. These status 
-/*	messages are of the form: "<junk><speed><junk>".
-/*	\fIagetty\fP assumes that the modem emits its status message at 
-/*	the same speed as specified with (the first) \fIbaud_rate\fP value
-/*	on the command line.
-/* .sp
-/*	Since the \fI-m\fP feature may fail on heavily-loaded systems, 
-/*	you still should enable BREAK processing by enumerating all 
-/*	expected baud rates on the command line.
-/* .TP
-/* -t timeout
-/*	Terminate if no user name could be read within \fItimeout\fP 
-/*	seconds. This option should probably not be used with hard-wired 
-/*	lines.
-/* SYSTEM V EXAMPLES
-/*	This section shows sample entries for the \fI/etc/inittab\fP file.
-/*
-/*	For a hard-wired line:
-/* .ti +5
-/*	t0:2:respawn:/etc/agetty ttyM0 9600
-/*
-/*	For a dial-in line with a 2400/1200/300 baud modem:
-/* .ti +5
-/*	t1:2:respawn:/etc/agetty -mt60 ttyM1 2400,1200,300
-/* SUNOS 4 EXAMPLES
-/* .ad
-/* .fi
-/*	This section show sample entries for the \fI/etc/ttytab\fP file.
-/*	Note that init(8) appends the port name to the command
-/*	specified in the inittab file.
-/*
-/*	For a hard-wired line:
-/* .ti +5
-/*	ttya  "/usr/etc/agetty 9600"  vt100  on local
-/*
-/*	For a dial-in line with a 2400/1200/300 baud modem:
-/* .ti +5
-/*	ttyb  "/usr/etc/agetty -mt60 2400,1200,300"  unknown  on modem
-/*
-/*	The latter also requires that the \fIDTR\fP and \fICD\fP modem 
-/*	control lines are enabled (see the eeprom(8) manual page).
-/* FILES
-/*	/etc/utmp, the system status file (System V only).
-/*	/etc/issue, printed before the login prompt (System V only).
-/*	/dev/console, problem reports (if syslog(3) is not used).
-/*	/etc/inittab (System V init(8) configuration file).
-/*	/etc/ttytab (SunOS 4 init(8) configuration file).
-/* BUGS
-/*	The baud-rate detection feature (the \fI-m\fP option) requires that
-/*	\fIagetty\fP be scheduled soon enough after completion of a dial-in
-/*	call (within 30 ms with modems that talk at 2400 baud). For robustness,
-/*	always use the \fI-m\fP option in combination with a multiple baud
-/*	rate command-line argument, so that BREAK processing is enabled.
-/*
-/*	The text in the /etc/issue file (System V only) and the login prompt
-/*	are always output with 7-bit characters and space parity.
-/*
-/*	The baud-rate detection feature (the \fI-m\fP option) requires that 
-/*	the modem emits its status message \fIafter\fP raising the DCD line.
-/* DIAGNOSTICS
-/*	Depending on how the program was configured, all diagnostics are
-/*	written to the console device or reported via the syslog(3) facility.
-/*	Error messages are produced if the \fIport\fP argument does not
-/*	specify a terminal device; if there is no /etc/utmp entry for the
-/*	current process (System V only); and so on.
-/* AUTHOR(S)
-/*	W.Z. Venema <wietse@wzv.win.tue.nl>
-/*	Eindhoven University of Technology
-/*	Department of Mathematics and Computer Science
-/*	Den Dolech 2, P.O. Box 513, 5600 MB Eindhoven, The Netherlands
-/* CREATION DATE
-/*	Sat Nov 25 22:51:05 MET 1989
-/* LAST MODIFICATION
-/*	91/09/01 23:22:00
-/* VERSION/RELEASE
-/*	1.29
-/*--*/
+/* agetty.c - another getty program for Linux. By W. Z. Venema 1989
+   Ported to Linux by Peter Orbaek <poe@daimi.aau.dk>
+   This program in freely distributable. The entire man-page used to
+   be here. Now read the real man-page agetty.8 instead.
+*/
 
 #ifndef	lint
 char sccsid[] = "@(#) agetty.c 1.29 9/1/91 23:22:00";
 #endif
 
 #include <stdio.h>
-
+#include <unistd.h>
+#include <stdlib.h>
+#include <sys/ioctl.h>
 #include <termio.h>
 #include <signal.h>
 #include <errno.h>
@@ -175,11 +23,12 @@ char sccsid[] = "@(#) agetty.c 1.29 9/1/91 23:22:00";
 #include <utmp.h>
 #include <getopt.h>
 #include <memory.h>
+#include <sys/file.h>
 
 #ifdef linux
 #include "pathnames.h"
 #include <sys/param.h>
-#undef USE_SYSLOG
+#define USE_SYSLOG
 #endif
 
  /* If USE_SYSLOG is undefined all diagnostics go directly to /dev/console. */
@@ -189,8 +38,6 @@ char sccsid[] = "@(#) agetty.c 1.29 9/1/91 23:22:00";
 extern void closelog();
 #endif
 
-extern void exit();
-
  /*
   * Some heuristics to find out what environment we are in: if it is not
   * System V, assume it is SunOS 4.
@@ -290,6 +137,7 @@ struct options {
 #define	F_PARSE		(1<<0)		/* process modem status messages */
 #define	F_ISSUE		(1<<1)		/* display /etc/issue */
 #define	F_RTSCTS	(1<<2)		/* enable RTS/CTS flow control */
+#define F_LOCAL		(1<<3)		/* force local */
 
 /* Storage for things detected while the login name was read. */
 
@@ -311,6 +159,23 @@ struct chardata init_chardata = {
     0,					/* always filled in at runtime */
 };
 
+#define P_(s) ()
+void parse_args P_((int argc, char **argv, struct options *op));
+void parse_speeds P_((struct options *op, char *arg));
+void update_utmp P_((char *line));
+void open_tty P_((char *tty, struct termio *tp, int local));
+void termio_init P_((struct termio *tp, int speed, int local));
+void auto_baud P_((struct termio *tp));
+void do_prompt P_((struct options *op, struct termio *tp));
+void next_speed P_((struct termio *tp, struct options *op));
+char *get_logname P_((struct options *op, struct chardata *cp, struct termio *tp));
+void termio_final P_((struct options *op, struct termio *tp, struct chardata *cp));
+int caps_lock P_((char *s));
+int bcode P_((char *s));
+void usage P_((void));
+void error P_((int va_alist));
+#undef P_
+
 /* The following is used for understandable diagnostics. */
 
 char *progname;
@@ -323,9 +188,10 @@ FILE *dbf;
 #define debug(s) /* nothing */
 #endif
 
+int
 main(argc, argv)
-int     argc;
-char  **argv;
+     int     argc;
+     char  **argv;
 {
     char   *logname;			/* login name, given to /bin/login */
     char   *get_logname();
@@ -372,7 +238,7 @@ char  **argv;
 #endif
 
     /* Open the tty as standard { input, output, error }. */
-    open_tty(options.tty, &termio);
+    open_tty(options.tty, &termio, options.flags & F_LOCAL);
 
 #ifdef linux
 	{
@@ -384,7 +250,7 @@ char  **argv;
 #endif
     /* Initialize the termio settings (raw mode, eight-bit, blocking i/o). */
 
-    termio_init(&termio, options.speeds[FIRST_SPEED]);
+    termio_init(&termio, options.speeds[FIRST_SPEED], options.flags & F_LOCAL);
 	
     /* Optionally detect the baud rate from the modem status message. */
 
@@ -416,24 +282,28 @@ char  **argv;
 
     /* Let the login program take care of password validation. */
 
-    (void) execl(options.login, options.login, logname, (char *) 0);
+    (void) execl(options.login, options.login, "--", logname, (char *) 0);
     error("%s: can't exec %s: %m", options.tty, options.login);
-    /* NOTREACHED */
+    exit(0);  /* quiet GCC */
 }
 
 /* parse-args - parse command-line arguments */
 
+void
 parse_args(argc, argv, op)
-int     argc;
-char  **argv;
-struct options *op;
+     int     argc;
+     char  **argv;
+     struct options *op;
 {
     extern char *optarg;		/* getopt */
     extern int optind;			/* getopt */
     int     c;
 
-    while (isascii(c = getopt(argc, argv, "hil:mt:"))) {
+    while (isascii(c = getopt(argc, argv, "Lhil:mt:"))) {
 	switch (c) {
+	case 'L':				/* force local */
+ 	    op->flags |= F_LOCAL;
+ 	    break;
 	case 'h':				/* enable h/w flow control */
 	    op->flags |= F_RTSCTS;
 	    break;
@@ -458,27 +328,25 @@ struct options *op;
     if (argc != optind + 2)			/* check parameter count */
 	usage();
 
-#ifdef linux
-    /* while we use a BSD like init and ttytab */
-    parse_speeds(op, argv[optind++]);		/* baud rate(s) */
-    op->tty = argv[optind];			/* tty name */
-#else    
-#ifdef	SYSV_STYLE
-    op->tty = argv[optind++];			/* tty name */
-    parse_speeds(op, argv[optind]);		/* baud rate(s) */
-#else
-    parse_speeds(op, argv[optind++]);		/* baud rate(s) */
-    op->tty = argv[optind];			/* tty name */
-#endif
-#endif /* linux */
-	debug("exiting parseargs\n");
+    /* we loosen up a bit and accept both "baudrate tty" and "tty baudrate" */
+    if('0' <= argv[optind][0] && argv[optind][0] <= '9') {
+	/* a number first, assume it's a speed (BSD style) */
+	parse_speeds(op, argv[optind++]);	/* baud rate(s) */
+	op->tty = argv[optind];			/* tty name */
+    } else {
+	op->tty = argv[optind++];		/* tty name */
+	parse_speeds(op, argv[optind]);		/* baud rate(s) */
+    }
+
+    debug("exiting parseargs\n");
 }
 
 /* parse_speeds - parse alternate baud rates */
 
+void
 parse_speeds(op, arg)
-struct options *op;
-char   *arg;
+     struct options *op;
+     char   *arg;
 {
     char   *strtok();
     char   *cp;
@@ -496,9 +364,9 @@ char   *arg;
 #ifdef	SYSV_STYLE
 
 /* update_utmp - update our utmp entry */
-
+void
 update_utmp(line)
-char   *line;
+     char   *line;
 {
     struct utmp ut;
     long    ut_size = sizeof(ut);	/* avoid nonsense */
@@ -531,8 +399,10 @@ char   *line;
 	endutent();
 
         if((ut_fd = open(_PATH_WTMP, O_APPEND|O_WRONLY)) >= 0) {
-		write(ut_fd, &ut, sizeof(ut));
-		close(ut_fd);
+	    flock(ut_fd, LOCK_EX);
+	    write(ut_fd, &ut, sizeof(ut));
+	    flock(ut_fd, LOCK_UN);
+	    close(ut_fd);
 	}
 #else
     if ((ut_fd = open(UTMP_FILE, 2)) < 0) {
@@ -558,10 +428,11 @@ char   *line;
 #endif
 
 /* open_tty - set up tty as standard { input, output, error } */
-
-open_tty(tty, tp)
-char   *tty;
-struct termio *tp;
+void
+open_tty(tty, tp, local)
+     char   *tty;
+     struct termio *tp;
+     int    local;
 {
     /* Get rid of the present standard { output, error} if any. */
 
@@ -588,7 +459,7 @@ struct termio *tp;
 	(void) close(0);
 	errno = 0;				/* ignore close(2) errors */
 
-	if (open(tty, O_RDWR, 0) != 0)
+	if (open(tty, (local ? O_RDWR|O_NONBLOCK : O_RDWR), 0) != 0)
 	    error("/dev/%s: cannot open as standard input: %m", tty);
 
     } else {
@@ -636,9 +507,11 @@ struct termio *tp;
 char gbuf[1024];
 char area[1024];
 
-termio_init(tp, speed)
-struct termio *tp;
-int     speed;
+void
+termio_init(tp, speed, local)
+     struct termio *tp;
+     int     speed;
+     int	local;
 {
 
     /*
@@ -653,17 +526,24 @@ int     speed;
 #endif
 
     tp->c_cflag = CS8 | HUPCL | CREAD | speed;
+    if (local) {
+	tp->c_cflag |= CLOCAL;
+    }
+
     tp->c_iflag = tp->c_lflag = tp->c_oflag = tp->c_line = 0;
     tp->c_cc[VMIN] = 1;
     tp->c_cc[VTIME] = 0;
     (void) ioctl(0, TCSETA, tp);
+    if (local) {
+	(void) fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NDELAY);
+    }
     debug("term_io 2\n");
 }
 
 /* auto_baud - extract baud rate from modem status message */
-
+void
 auto_baud(tp)
-struct termio *tp;
+     struct termio *tp;
 {
     int     speed;
     int     vmin;
@@ -724,15 +604,14 @@ struct termio *tp;
 }
 
 /* do_prompt - show login prompt, optionally preceded by /etc/issue contents */
-
+void
 do_prompt(op, tp)
-struct options *op;
-struct termio *tp;
+     struct options *op;
+     struct termio *tp;
 {
 #ifdef	ISSUE
     FILE    *fd;
     int     oflag;
-    int     n;
     char    c;
     struct utsname uts;
 
@@ -775,6 +654,10 @@ struct termio *tp;
 		    (void) printf ("%s", uts.machine);
 		    break;
 
+		  case 'o':
+		    (void) printf ("%s", uts.domainname);
+		    break;
+
 		  case 'd':
 		  case 't':
 		    {
@@ -843,10 +726,10 @@ struct termio *tp;
 }
 
 /* next_speed - select next baud rate */
-
+void
 next_speed(tp, op)
-struct termio *tp;
-struct options *op;
+     struct termio *tp;
+     struct options *op;
 {
     static int baud_index = FIRST_SPEED;/* current speed index */
 
@@ -859,9 +742,9 @@ struct options *op;
 /* get_logname - get user name, establish parity, speed, erase, kill, eol */
 
 char   *get_logname(op, cp, tp)
-struct options *op;
-struct chardata *cp;
-struct termio *tp;
+     struct options *op;
+     struct chardata *cp;
+     struct termio *tp;
 {
     char    logname[BUFSIZ];
     char   *bp;
@@ -968,16 +851,17 @@ struct termio *tp;
 }
 
 /* termio_final - set the final tty mode bits */
-
+void
 termio_final(op, tp, cp)
-struct options *op;
-struct termio *tp;
-struct chardata *cp;
+     struct options *op;
+     struct termio *tp;
+     struct chardata *cp;
 {
     /* General terminal-independent stuff. */
 
     tp->c_iflag |= IXON | IXOFF;		/* 2-way flow control */
-    tp->c_lflag |= ICANON | ISIG | ECHO | ECHOKE | ECHOCTL | ECHOPRT;
+    tp->c_lflag |= ICANON | ISIG | ECHO | ECHOE | ECHOK| ECHOKE;
+      /* no longer| ECHOCTL | ECHOPRT*/
     tp->c_oflag |= OPOST;
     /* tp->c_cflag = 0; */
     tp->c_cc[VINTR] = DEF_INTR;			/* default interrupt */
@@ -1037,9 +921,9 @@ struct chardata *cp;
 }
 
 /* caps_lock - string contains upper case without lower case */
-
+int
 caps_lock(s)
-char   *s;
+     char   *s;
 {
     int     capslock;
 
@@ -1053,9 +937,9 @@ char   *s;
 }
 
 /* bcode - convert speed string to speed code; return 0 on failure */
-
+int
 bcode(s)
-char   *s;
+     char   *s;
 {
     struct Speedtab {
 	long    speed;
@@ -1090,7 +974,6 @@ char   *s;
 	0, 0,
     };
     struct Speedtab *sp;
-    long    atol();
     long    speed = atol(s);
 
     for (sp = speedtab; sp->speed; sp++)
@@ -1101,6 +984,7 @@ char   *s;
 
 /* usage - explain */
 
+void
 usage()
 {
 #if defined(SYSV_STYLE) && !defined(linux)
@@ -1119,9 +1003,9 @@ usage()
 #define	str2cpy(b,s1,s2)	strcat(strcpy(b,s1),s2)
 
 /* VARARGS */
-
+void
 error(va_alist)
-va_dcl
+     va_dcl
 {
     va_list ap;
     char   *fmt;
