@@ -104,13 +104,6 @@ static char sccsid[] = "@(#)login.c	5.40 (Berkeley) 5/9/89";
 #define index strchr
 #define rindex strrchr
 #include <sys/ioctl.h>
-
-#ifdef TESTING
-#include "utmp.h"
-#else
-#include <utmp.h>
-#endif
-
 #include <signal.h>
 #include <errno.h>
 #include <grp.h>
@@ -120,6 +113,16 @@ static char sccsid[] = "@(#)login.c	5.40 (Berkeley) 5/9/89";
 #include <stdio.h>
 #include <string.h>
 #include <sys/syslog.h>
+#include <sys/sysmacros.h>
+#ifdef TESTING
+#  include "utmp.h"
+#else
+#  include <utmp.h>
+#endif
+
+#ifdef SHADOW_PWD
+#include <shadow.h>
+#endif
 
 #ifndef linux
 #include <tzfile.h>
@@ -134,6 +137,22 @@ struct  lastlog
 
 #include "pathnames.h"
 
+#define P_(s) ()
+void opentty P_((const char *tty));
+void getloginname P_((void));
+void timedout P_((void));
+int rootterm P_((char *ttyn));
+void motd P_((void));
+void sigint P_((void));
+void checknologin P_((void));
+void dolastlog P_((int quiet));
+void badlogin P_((char *name));
+char *stypeof P_((char *ttyid));
+void checktty P_((char *user, char *tty));
+void getstr P_((char *buf, int cnt, char *err));
+void sleepexit P_((int eval));
+#undef P_
+
 #ifdef	KERBEROS
 #include <kerberos/krb.h>
 #include <sys/termios.h>
@@ -181,7 +200,8 @@ char *months[] =
 	  "Sep", "Oct", "Nov", "Dec" };
 
 /* provided by Linus Torvalds 16-Feb-93 */
-void opentty(const char * tty)
+void 
+opentty(const char * tty)
 {
     int i;
     int fd = open(tty, O_RDWR);
@@ -194,7 +214,7 @@ void opentty(const char * tty)
       close(fd);
 }
 
-
+int
 main(argc, argv)
 	int argc;
 	char **argv;
@@ -210,10 +230,10 @@ main(argc, argv)
 	int quietlog, passwd_req, ioctlval;
 	char *domain, *salt, *ttyn, *pp;
 	char tbuf[MAXPATHLEN + 2], tname[sizeof(_PATH_TTY) + 10];
-	char *ctime(), *ttyname(), *stypeof(), *crypt();
+	char *ctime(), *ttyname(), *stypeof();
 	time_t time();
 	void timedout();
-	char termenv[20];
+	char *termenv;
 
 #ifdef linux
 	char tmp[100];
@@ -225,8 +245,9 @@ main(argc, argv)
 	(void)alarm((unsigned int)timeout);
 	(void)signal(SIGQUIT, SIG_IGN);
 	(void)signal(SIGINT, SIG_IGN);
-#ifndef linux
+
 	(void)setpriority(PRIO_PROCESS, 0, 0);
+#ifdef HAVE_QUOTA
 	(void)quota(Q_SETUID, 0, 0, 0);
 #endif
 
@@ -349,17 +370,9 @@ main(argc, argv)
 			fflag = 0;
 			getloginname();
 		}
-		/*
-		 * Note if trying multiple user names;
-		 * log failures for previous user name,
-		 * but don't bother logging one failure
-		 * for nonexistent name (mistyped username).
-		 */
-		if (failures && strcmp(tbuf, username)) {
-			if (failures > (pwd ? 0 : 1))
-				badlogin(tbuf);
-			failures = 0;
-		}
+
+		checktty(username, tty);
+
 		(void)strcpy(tbuf, username);
 		if (pwd = getpwnam(username))
 			salt = pwd->pw_passwd;
@@ -408,14 +421,10 @@ main(argc, argv)
 		if (!passwd_req || (pwd && !*pwd->pw_passwd))
 			break;
 
-#ifndef linux
 		setpriority(PRIO_PROCESS, 0, -4);
-#endif
 		pp = getpass("Password: ");
 		p = crypt(pp, salt);
-#ifndef linux
 		setpriority(PRIO_PROCESS, 0, 0);
-#endif
 
 #ifdef	KERBEROS
 
@@ -448,13 +457,11 @@ main(argc, argv)
 
 		(void)printf("Login incorrect\n");
 		failures++;
+		badlogin(username); /* log ALL bad logins */
+
 		/* we allow 10 tries, but after 3 we start backing off */
 		if (++cnt > 3) {
 			if (cnt >= 10) {
-				badlogin(username);
-#ifndef linux
-				(void)ioctl(0, TIOCHPCL, (struct sgttyb *)NULL);
-#endif
 				sleepexit(1);
 			}
 			sleep((unsigned int)((cnt - 3) * 5));
@@ -464,10 +471,7 @@ main(argc, argv)
 	/* committed to login -- turn off timeout */
 	(void)alarm((unsigned int)0);
 
-	/* paranoia... */
-	endpwent();
-
-#ifndef linux
+#ifdef HAVE_QUOTA
 	if (quota(Q_SETUID, pwd->pw_uid, 0, 0) < 0 && errno != EINVAL) {
 		switch(errno) {
 		case EUSERS:
@@ -485,15 +489,32 @@ main(argc, argv)
 	}
 #endif
 
-	if (chdir(pwd->pw_dir) < 0) {
-		(void)printf("No directory %s!\n", pwd->pw_dir);
-		if (chdir("/"))
-			exit(0);
-		pwd->pw_dir = "/";
-		(void)printf("Logging in with home = \"/\".\n");
-	}
+	/* paranoia... */
+	endpwent();
 
-	quietlog = access(_PATH_HUSHLOGIN, 004) == 0;
+	/* This requires some explanation: As root we may not be able to
+	   read the directory of the user if it is on an NFS mounted
+	   filesystem. We temporarily set our effective uid to the user-uid
+	   making sure that we keep root privs. in the real uid. 
+
+	   A portable solution would require a fork(), but we rely on Linux
+	   having the BSD setreuid() */
+
+	{
+	    char tmpstr[MAXPATHLEN];
+	    uid_t ruid = getuid();
+	    gid_t egid = getegid();
+
+	    strncpy(tmpstr, pwd->pw_dir, MAXPATHLEN-12);
+	    strncat(tmpstr, ("/" _PATH_HUSHLOGIN), MAXPATHLEN);
+
+	    setregid(-1, pwd->pw_gid);
+	    setreuid(0, pwd->pw_uid);
+	    quietlog = (access(tmpstr, R_OK) == 0);
+	    setuid(0); /* setreuid doesn't do it alone! */
+	    setreuid(ruid, 0);
+	    setregid(-1, egid);
+	}
 
 #ifndef linux
 #ifdef KERBEROS
@@ -563,7 +584,9 @@ main(argc, argv)
 		endutent();
 		
 		if((wtmp = open(_PATH_WTMP, O_APPEND|O_WRONLY)) >= 0) {
+		        flock(wtmp, LOCK_EX);
 			write(wtmp, (char *)&ut, sizeof(ut));
+		        flock(wtmp, LOCK_UN);
 			close(wtmp);
 		}
 	}
@@ -581,16 +604,13 @@ main(argc, argv)
 #endif
 	(void)chown(ttyn, pwd->pw_uid,
 	    (gr = getgrnam(TTYGRPNAME)) ? gr->gr_gid : pwd->pw_gid);
-#ifndef linux
-	(void)chmod(ttyn, 0620);
-#else
+
 	(void)chmod(ttyn, 0622);
-#endif
 	(void)setgid(pwd->pw_gid);
 
 	initgroups(username, pwd->pw_gid);
 
-#ifndef linux
+#ifdef HAVE_QUOTA
 	quota(Q_DOWARN, pwd->pw_uid, (dev_t)-1, 0);
 #endif
 
@@ -608,10 +628,8 @@ main(argc, argv)
 	{
 		char *ep;
 		
-		if(ep = getenv("TERM"))
-			(void)strncpy(termenv, ep, sizeof(termenv) - 1);
-		else
-			(void)strcpy(termenv, "dumb");
+		if(!((ep = getenv("TERM")) && (termenv = strdup(ep))))
+		  termenv = "dumb";
 	}
 
 	/* destroy environment unless user has requested preservation */
@@ -680,13 +698,26 @@ main(argc, argv)
 	    p + 1 : pwd->pw_shell);
 
 	/* discard permissions last so can't get killed and drop core */
-	(void)setuid(pwd->pw_uid);
+	if(setuid(pwd->pw_uid) < 0 && pwd->pw_uid) {
+	    syslog(LOG_ALERT, "setuid() failed");
+	    exit(1);
+	}
+
+	/* wait until here to change directory! */
+	if (chdir(pwd->pw_dir) < 0) {
+		(void)printf("No directory %s!\n", pwd->pw_dir);
+		if (chdir("/"))
+			exit(0);
+		pwd->pw_dir = "/";
+		(void)printf("Logging in with home = \"/\".\n");
+	}
 
 	execlp(pwd->pw_shell, tbuf, (char *)0);
 	(void)fprintf(stderr, "login: no shell: %s.\n", strerror(errno));
 	exit(0);
 }
 
+void
 getloginname()
 {
 	register int ch;
@@ -728,6 +759,7 @@ void timedout()
 	exit(0);
 }
 
+int
 rootterm(ttyn)
 	char *ttyn;
 #ifndef linux
@@ -768,6 +800,7 @@ rootterm(ttyn)
 
 jmp_buf motdinterrupt;
 
+void
 motd()
 {
 	register int fd, nchars;
@@ -789,6 +822,7 @@ void sigint()
 	longjmp(motdinterrupt, 1);
 }
 
+void
 checknologin()
 {
 	register int fd, nchars;
@@ -801,6 +835,7 @@ checknologin()
 	}
 }
 
+void
 dolastlog(quiet)
 	int quiet;
 {
@@ -814,12 +849,13 @@ dolastlog(quiet)
 			    ll.ll_time != 0) {
 				(void)printf("Last login: %.*s ",
 				    24-5, (char *)ctime(&ll.ll_time));
+
 				if (*ll.ll_host != '\0')
-					(void)printf("from %.*s\n",
-					    sizeof(ll.ll_host), ll.ll_host);
+				  printf("from %.*s\n",
+					 (int)sizeof(ll.ll_host), ll.ll_host);
 				else
-					(void)printf("on %.*s\n",
-					    sizeof(ll.ll_line), ll.ll_line);
+				  printf("on %.*s\n",
+					 (int)sizeof(ll.ll_line), ll.ll_line);
 			}
 			(void)lseek(fd, (off_t)pwd->pw_uid * sizeof(ll), L_SET);
 		}
@@ -833,6 +869,7 @@ dolastlog(quiet)
 	}
 }
 
+void
 badlogin(name)
 	char *name;
 {
@@ -861,6 +898,63 @@ stypeof(ttyid)
 }
 #endif 
 
+void
+checktty(user, tty)
+     char *user;
+     char *tty;
+{
+    FILE *f;
+    char buf[256];
+    char *ptr;
+    char devname[50];
+    struct stat stb;
+
+    /* no /etc/usertty, default to allow access */
+    if(!(f = fopen(_PATH_USERTTY, "r"))) return;
+
+    while(fgets(buf, 255, f)) {
+
+	/* strip comments */
+	for(ptr = buf; ptr < buf + 256; ptr++) 
+	  if(*ptr == '#') *ptr = 0;
+
+	strtok(buf, " \t");
+	if(strncmp(user, buf, 8) == 0) {
+	    while((ptr = strtok(NULL, "\t\n "))) {
+		if(strncmp(tty, ptr, 10) == 0) {
+		    fclose(f);
+		    return;
+		}
+		if(strcmp("PTY", ptr) == 0) {
+#ifdef linux
+		    sprintf(devname, "/dev/%s", ptr);
+		    /* VERY linux dependent, recognize PTY as alias
+		       for all pseudo tty's */
+		    if((stat(devname, &stb) >= 0)
+		       && major(stb.st_rdev) == 4 
+		       && minor(stb.st_rdev) >= 192) {
+			fclose(f);
+			return;
+		    }
+#endif
+		}
+	    }
+	    /* if we get here, /etc/usertty exists, there's a line
+	       beginning with our username, but it doesn't contain the
+	       name of the tty where the user is trying to log in.
+	       So deny access! */
+	    fclose(f);
+	    printf("Login on %s denied.\n", tty);
+	    badlogin(user);
+	    sleepexit(1);
+	}
+    }
+    fclose(f);
+    /* users not mentioned in /etc/usertty are by default allowed access
+       on all tty's */
+}
+
+void
 getstr(buf, cnt, err)
 	char *buf, *err;
 	int cnt;
@@ -878,6 +972,7 @@ getstr(buf, cnt, err)
 	} while (ch);
 }
 
+void
 sleepexit(eval)
 	int eval;
 {
