@@ -9,6 +9,7 @@
 #include <pwd.h>
 #include <unistd.h>
 #include "pathnames.h"
+#include <sys/param.h>
 
 #ifdef SHADOW_PWD
 #include <shadow.h>
@@ -20,6 +21,9 @@ char message[BSIZ];
 char *prog;
 int broadcast;
 
+char hostname[MAXHOSTNAMELEN];
+char mytty[8];	/* 8 should do it 8-) - st */
+
 void usage()
 {
 	fprintf(stderr, "Usage: %s user ...\n", prog);
@@ -42,6 +46,14 @@ void write_line(struct utmp *ut)
 	struct passwd *pw;
 	extern char *getlogin();
 
+	time_t time_now;
+	char *timestamp;
+	
+	/* get time now for message timestamp... */
+	time(&time_now);
+	timestamp=ctime(&time_now);
+	timestamp+=11;
+	
 	(void) strcat(term, ut->ut_line);
 
 	if(!(user = getlogin())) {
@@ -50,9 +62,9 @@ void write_line(struct utmp *ut)
 	}
 
 	if(f = fopen(term, "w")) {
-		fprintf(f, "\r\n%sessage from %s:\r\n", 
+		fprintf(f, "\r\n\007\007\007%sessage from %s@%s on %s at %5.5s ...\r\n", 
 			broadcast ? "Broadcast m" : "M",
-			user);
+			user, hostname, mytty, timestamp);
 		fwrite(message, 1, strlen(message), f);
 		fputs("\r\n", f);
 		fclose(f);
@@ -99,6 +111,9 @@ int main(argc, argv)
 	int i, j;
 	char *p;
 	
+	gethostname(hostname, MAXHOSTNAMELEN);
+	strcpy(mytty,(strrchr(ttyname(0), '/'))+1);
+	
 	prog = argv[0];
 	if(p = strrchr(argv[0], '/')) prog = p + 1;
 
